from datetime import datetime
from typing import Dict
import pandas as pd
import abc
import threading
from loguru import logger as log
from pathlib import Path

class Writer(object):

    # TODO: get chunksize from config
    def __init__(self):
        self.schema = [
            'source',
            'created_at',
            'lang',
            'reply_settings',
            'referenced_tweets',
            'possibly_sensitive',
            'author_id',
            'id',
            'text',
            'conversation_id',
            'public_metrics.retweet_count',
            'public_metrics.reply_count',
            'public_metrics.like_count',
            'public_metrics.quote_count',
            'entities.mentions',
            'in_reply_to_user_id',
            'entities.urls',
            'entities.hashtags',
            'attachments.media_keys',
            'context_annotations'
        ]
        self.buffer = pd.DataFrame()
        # if (self.mp == True):
        self.lock = threading.Lock()
        self.path = Path(f'{datetime.strftime(datetime.now(), "%Y%m%d_%H%M")}_twacapic.csv')
        # create output file
        with self.path.open('w') as file:
            file.writelines(f'{",".join(self.schema)}\n')
        self.chunkSize = 100

    def __del__(self):
        # flush buffer before destruction
        self.persist(len(self.buffer))

    def write(self, df: pd.DataFrame):
        # if (self.mp == True):
        with self.lock:
            self.buffer = pd.concat([self.buffer, df], ignore_index=True, sort=False)
            log.info(f'Buffer now contains {len(self.buffer)} records for thread {id}.')

            # Call write persistance loop to empty buffer
            while (len(self.buffer) > self.chunkSize):
                self.persist(len(self.buffer))
    # else:
        #     self.buffer = pd.concat([self.buffer, df], ignore_index=True, sort=False)
        #     log.info(f'Buffer now contains {len(self.buffer)} records for thread {id}.')

        #     # Call write persistance loop to empty buffer
        #     while (len(self.buffer) > self.chunkSize):
        #         self.persist(len(self.buffer))

    @abc.abstractmethod
    def persist(self, chunkSize: int) -> None:
        return None



from pathlib import Path
import abc
from typing import Dict

class Reader(object):
    
    config: Dict
    workingPath: Path
    __setup = False

    @staticmethod
    def setup(path: Path) -> None:
        """
        This hook is called by programm start up in order to load additional resources and/or configuration from the working folder.
        Implementation here serves solely as an interface and guideline for implmentation in subclasses.

        Never call this static method from your class definition to set `__setup` to true!
        """
        raise "You shall not call me, understand!"
        Reader.workingPath = Path
        Reader.config = read(path)
        Reader.__setup = True
        return

    def __init__(self, path: Path):
        self.path = path

    @abc.abstractmethod
    def read(self):
        return



from pathlib import Path
from json import load
from typing import Dict
import pandas as pd
from loguru import logger as log
from .Reader import Reader

class TwacapicReader(Reader):

    @staticmethod
    def setup(path: Path) -> None:
        # what did I want do here?
        return
   
    @staticmethod
    def getSchema() -> list(str):
        if (TwacapicReader.__setup == False):
            raise 'Tryed to use TwacapicReader without setting it up first'
        if (TwacapicReader.config is None):
            raise 'Tryed to use TwacapicReader without proper configuration'
        return [""]

    def __init__(self, path: Path, property: str = 'data'):
        super().__init__(path)
        self.property = property
    
    def read(self):
        schema = TwacapicReader.getSchema()
        data = None
        with self.path.open('r') as file:
            try:
                data = load(file)['data']
            except Exception as e:
                log.error(e)
        data = pd.json_normalize(data)
        log.debug(f'Found {len(data)} records in {str(self.path)}')
        return data


from pathlib import Path
import importlib
from typing import Dict, Literal
import click
import sys
import yaml
from multiprocessing import cpu_count
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor
from threading import get_ident, active_count
from loguru import logger as log
from .read import read

@click.command()
@click.argument('input', type=click.Path(exists=True))
@click.option('--pattern', default='*.json', help='file extension to look for')
@click.option('--host', default='localhost', help='Host address of the dabase too write to.')
@click.option('--port', default='5432', help='Host port of the dabase-server.')
@click.option('--dbname', default='dabapushed', help='Name of the dabase too write to.')
@click.option('--n_workers', default=cpu_count, help="Number of worker threads to read/write data")
@click.option('--reader', default='Twacapic', help='python class to read stuff (see docs). Possible values: Twacapic')
@click.option('--writer', default='CSV',help='python class to write stuff (see docs)')
@click.option('--recursive', '-r', help='should the reader recurse in subdirectories? Default: TRUE.', is_flag=True)
@click.option('--logfile', help='file to logi in (optional)')
@click.option('--loglevel', default='INFO', help='the level to log, yk')
def run(
    input: str,
    pattern: str,
    reader: str,
    writer: str,
    host: str,
    port: str,
    dbname: str,
    n_workers: str,
    recursive: bool,
    logfile: str,
    loglevel: str
):
    # load config
    config = yaml.load(Path('config.yml').open(), Loader=yaml.FullLoader)
    
    log.remove()
    log.add(sys.stdout, level=loglevel)
    log.debug('Using this configuration', config)
    # Validate all inputs
    if logfile is not None:
        log.add(logfile, level=loglevel, rotation="64 MB")
        log.add(sys.stderr, level='ERROR')

    log.add('errors.log', level='ERROR')
    log.add('warnings.log', level='WARNING')

    log.info(f'{input}**/*.{pattern} will be written to {host}:{port}/{dbname} with {n_workers} parallel threads')

    ReaderClass = None
    # Load the reader/writer:
    ReaderClass = load_from_config(reader, 'reader', config)
    WriterClass = load_from_config(writer, 'writer', config)

    # Fire up the engines and find in all matching files
    files = read(input, pattern, recursive=recursive)

    # Get a Writer
    writerInstance = WriterClass()

    def proop(thing: Path) -> any:
        readerInstance = ReaderClass(thing)
        tid = get_ident()
        log.debug(f'Reading {thing} with Thread ID: {tid}')
        return writerInstance.write(readerInstance.read())
        
    with ThreadPoolExecutor() as executor:
        # log.debug(f'Starting {active_count()} threads.')
        lines = executor.map(proop, files)
        # log.info(f'Read a total of {sum(lines)} records')

def load_from_config(name: str, key: str, config: Dict):
    if (name is not None and name in config['plugins'][key].keys()):
        log.debug(f'Using this {key}: {name}')
        try:
            moduleName = config['plugins'][key][name]['moduleName']
            className = config['plugins'][key][name]['className']
            ReaderClass = importlib.import_module(moduleName, package='dabapush').__getattribute__(className)
        except Exception as e:
            log.error(e)
        if (ReaderClass is not None):
            return ReaderClass
    else:
        log.warning(f'{key.upper()} {name} cannot be found')
    # start $n_workers workers to read the data
    # if JSON accecssor is given, apply it to each loaded file
    # writer(host, port, dbname)

if __name__ == '__main__':
    run()


from sys import path
from pathlib import Path
from typing import Generator
from loguru import logger as log

def read(path: str, pattern: str, recursive: bool) -> Generator[Path, None, None]:
    log.info(f'I will read here: {path}')
    
    _path = Path(path).resolve()
    return _path.rglob(pattern) if (recursive == True) else _path.glob(pattern)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dabapush', 'dabapush.Reader', 'dabapush.Writer']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.4.1,<6.0.0',
 'SQLAlchemy>=1.4.22,<2.0.0',
 'click>=8.0.0,<9.0.0',
 'loguru>=0.5.3,<0.6.0',
 'pandas>=1.3.1,<2.0.0']

entry_points = \
{'console_scripts': ['dabapush = dabapush.dabapush:run']}

setup_kwargs = {
    'name': 'dabapush',
    'version': '0.1.1',
    'description': '',
    'long_description': "# dabapush\n\nData Base pusher for social media data (Twitter for the beginning) – pre-alpha version\n\n## Developer Install\n\n1. Install [poetry](https://python-poetry.org/docs/#installation)\n2. Clone repository\n3. In the cloned repository's root directory run `poetry install`\n4. Run `poetry shell` to start development virtualenv\n5. Run `twacapic` to enter API keys. Ignore the IndexError.\n6. Run `pytest` to run all tests\n",
    'author': 'Felix Victor Münch',
    'author_email': 'f.muench@leibniz-hbi.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Leibniz-HBI/dabapush',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from loguru import logger as log
from pathlib import Path
from .Writer import Writer

class CSVWriter(Writer):
    
    def __init__(self):
        super().__init__()

    def persist(self, chunkSize: int):

        last_row = self.buffer.head(chunkSize)
        self.buffer.drop(last_row.index, inplace=True)

        log.info(f'Persisted {len(last_row)} records')
        
        with self.path.open('a') as file:
            last_row.replace(r'\n|\r', r'\\n', regex=True, inplace=True)
            last_row[self.schema].to_csv(file, index=False, header=False)
        return len(last_row)
